#
# scspell
# Copyright (C) 2009 Paul Pelzl
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 2, as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

"""scspell -- an interactive, conservative spell-checker for source code."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import argparse
import os
import re
import sys
import shutil
import uuid

try:
    import ConfigParser
except ImportError:
    # Python 3
    import configparser as ConfigParser

from . import _portable
from ._corpus import CorporaFile
from . import _util

from ._util import set_verbosity
from ._util import VERBOSITY_NORMAL
from ._util import VERBOSITY_MAX


assert set_verbosity
assert VERBOSITY_NORMAL is not None
assert VERBOSITY_MAX is not None


try:
    raw_input
except NameError:
    raw_input = input


__version__ = '2.1'

# Name of scspell.conf section header
CONFIG_SECTION = 'Settings'

# Size of context printed upon request
CONTEXT_SIZE = 4

# Subtokens shorter than 4 characters are likely to be abbreviations
LEN_THRESHOLD = 3

USER_DATA_DIR = _portable.get_data_dir('scspell')
DICT_DEFAULT_LOC = os.path.join(USER_DATA_DIR, 'dictionary.txt')
SCSPELL_DATA_DIR = os.path.normpath(
    os.path.join(os.path.dirname(__file__), 'data'))
SCSPELL_BUILTIN_DICT = os.path.join(SCSPELL_DATA_DIR, 'dictionary.txt')

SCSPELL_CONF = os.path.join(USER_DATA_DIR, 'scspell.conf')

# Treat anything alphanumeric as a token of interest, as long as it is not
# immediately preceded by a single backslash.  (The string "\ntext" should
# match on "text" rather than "ntext".)
C_ESCAPE_TOKEN_REGEX = re.compile(r'(?<![^\\]\\)\w+')

# \ is not a character escape in e.g. LaTeX
TOKEN_REGEX = re.compile(r'\w+')

# Hex digits will be treated as a special case, because they can look like
# word-like even though they are actually numeric
HEX_REGEX = re.compile(r'0x[0-9a-fA-F]+')

# We assume that tokens will be split using either underscores,
# digits, or camelCase conventions (or both)
US_REGEX = re.compile(r'[_\d]+')
CAMEL_WORD_REGEX = re.compile(r'([A-Z][a-z]*)')

# File-id specifiers take this form
FILE_ID_REGEX = re.compile(r'scspell-id:[ \t]*([a-zA-Z0-9_\-]+)')


class MatchDescriptor(object):

    """A MatchDescriptor captures the information necessary to represent a
    token matched within some source code."""

    def __init__(self, text, match_obj):
        self._data = text
        self._pos = match_obj.start()
        self._token = match_obj.group()
        self._context = None
        self._line_num = None

    def get_token(self):
        return self._token

    def get_string(self):
        """Get the entire string in which the match was found."""
        return self._data

    def get_ofs(self):
        """Get the offset within the string where the match is located."""
        return self._pos

    def get_prefix(self):
        """Get the string preceding this match."""
        return self._data[:self._pos]

    def get_remainder(self):
        """Get the string consisting of this match and all remaining
        characters."""
        return self._data[self._pos:]

    def get_context(self):
        """Compute the lines of context associated with this match, as a
        sequence of (line_num, line_string) pairs."""
        if self._context is not None:
            return self._context

        lines = self._data.split('\n')

        # Compute the byte offset of start of every line
        offsets = []
        for i in range(len(lines)):
            if i == 0:
                offsets.append(0)
            else:
                offsets.append(offsets[i - 1] + len(lines[i - 1]) + 1)

        # Compute the line number where the match is located
        for (i, ofs) in enumerate(offsets):
            if ofs > self._pos:
                self._line_num = i
                break
        if self._line_num is None:
            self._line_num = len(lines)

        # Compute the set of lines surrounding this line number
        self._context = [
            (i + 1, line.strip('\r\n'))for (i, line) in enumerate(lines)
            if (i + 1 - self._line_num) in
            range(-CONTEXT_SIZE // 2, CONTEXT_SIZE // 2 + 1)]
        return self._context

    def get_line_num(self):
        """Computes the line number of the match."""
        if self._line_num is None:
            self.get_context()
        return self._line_num


def make_unique(items):
    """Remove duplicate items from a list, while preserving list order."""
    seen = set()

    def first_occurrence(i):
        if i not in seen:
            seen.add(i)
            return True
        return False
    return [i for i in items if first_occurrence(i)]


def get_new_file_id():
    """Produce a new file ID string."""
    return str(uuid.uuid1())


def decompose_token(token):
    """Divide a token into a list of strings of letters.

    Tokens are divided by underscores and digits, and capital letters will
    begin new subtokens.

    :param token: string to be divided
    :returns: sequence of subtoken strings

    """
    us_parts = US_REGEX.split(token)
    if ''.join(us_parts).isupper():
        # This looks like a CONSTANT_DEFINE_OF_SOME_SORT
        subtokens = us_parts
    else:
        camelcase_parts = [
            CAMEL_WORD_REGEX.split(
                us_part) for us_part in us_parts]
        subtokens = sum(camelcase_parts, [])
    # This use of split() will create many empty strings
    return [st.lower() for st in subtokens if st != '']


def handle_new_filetype(extension, dicts):
    """Handle creation of a new file-type for the given extension.

    :returns: True if created, False if canceled.

    """
    while True:
        descr = raw_input("""\
            Enter a descriptive name for the programming language: """).strip()
        if descr == '':
            print("""\
            (Canceled.)\n""")
            return False

        if (':' in descr) or (';' in descr):
            print("""\
            Illegal characters in descriptive name.""")
            continue

        if descr in dicts.get_filetypes():
            print("""\
            That name is already in use.""")
            continue

        dicts.new_filetype(descr, [extension])
        return True


def handle_new_extension(ext, dicts):
    """Handle creation of a new file-type extension.

    :returns: True if new extension was registered, False if canceled.

    """
    print(("""\
            Extension "%s" is not registered.  With which programming language
            should "%s" be associated?""" % (ext, ext)))

    type_format = """\
               %3u: %s"""
    filetypes = dicts.get_filetypes()
    for i, ft in enumerate(filetypes):
        print(type_format % (i, ft))
    print(type_format % (len(filetypes), '(Create new language file-type)'))

    while True:
        selection = raw_input("""\
            Enter number of desired file-type: """)
        if selection == '':
            print("""\
            (Canceled.)\n""")
            return False

        try:
            selection = int(selection)
        except ValueError:
            continue
        if selection == len(filetypes):
            return handle_new_filetype(ext, dicts)
        elif selection >= 0 and selection < len(filetypes):
            dicts.register_extension(ext, filetypes[selection])
            return True


def build_add_prompt(offer_p, offer_f, offer_N):
    """Build a prompt for adding a word to a dictionary.

    :param offer_p: offer a (p)rogramming language dictionary
    :type offer_p: bool
    :param offer_f: offer a (f)ile-specific dictionary
    :type offer_f: bool
    :param offer_N: offer to create a (N)ew file-specific dictionary
    :type offer_p: bool
    :returns: prompt string

    """

    prompt = """\
      Subtoken '%s':
         (b)ack, (i)gnore, add to (n)atural language dictionary"""
    if offer_p:
        prompt += """, add to
         (p)rogramming language dictionary"""
    if offer_f:
        prompt += """, add to
         (f)ile-specific dictionary"""
    if offer_N:
        prompt += """, add to
         (N)ew file-specific dictionary"""
    prompt += """ [i]"""
    return prompt


def handle_add(unmatched_subtokens, filename, fq_filename, file_id_ref, dicts):
    """Handle addition of one or more subtokens to a dictionary.

    :param unmatched_subtokens: sequence of subtokens, each of which failed
           spell check
    :param filename: name of file containing the token
    :param fq_filename: fully-qualified filename
    :param file_id: unique identifier for current file: [string or None]
    :param dicts: dictionary set against which to perform matching
    :type  dicts: CorporaFile
    :returns: True if subtokens were handled, False if canceled

    """
    (_, ext) = os.path.splitext(filename.lower())

    prompt = None

    for subtoken in unmatched_subtokens:
        while True:
            if prompt is None:
                file_id = file_id_ref[0]
                offer_p = (ext != '')
                offer_f = (file_id is not None)
                offer_N = dicts._relative_to is not None and not offer_f
                prompt = build_add_prompt(offer_p, offer_f, offer_N)

            print(prompt % subtoken)
            ch = _portable.getch()
            if ch in (_portable.CTRL_C, _portable.CTRL_D, _portable.CTRL_Z):
                sys.exit(2)
            elif ch == 'b':
                print("""\
         (Canceled.)\n""")
                return False
            elif ch in ('i', '\r', '\n'):
                break
            elif offer_p and ch == 'p':
                if dicts.add_by_extension(subtoken, ext):
                    break
                else:
                    if handle_new_extension(ext, dicts) and \
                            dicts.add_by_extension(subtoken, ext):
                        break
            elif ch == 'n':
                dicts.add_natural(subtoken)
                break
            elif offer_f and (ch == 'f'):
                dicts.add_by_file_id(subtoken, file_id)
                break
            elif offer_N and (ch == 'N'):
                file_id = get_new_file_id()
                file_id_ref[0] = file_id
                print('New file ID {0} for {1}'.format(file_id, filename),
                      file=sys.stderr)
                dicts.new_file_and_file_id(fq_filename, file_id)
                dicts.add_by_file_id(subtoken, file_id)
                prompt = None  # reselect prompt now that file_id is not None
                break
    return True


def handle_failed_check_interactively(
        match_desc, filename, fq_filename, file_id_ref,
        unmatched_subtokens, dicts, ignores):
    """Handle a token which failed the spell check operation.

    :param match_desc: description of the token matching instance
    :type  match_desc: MatchDescriptor
    :param filename: name of file containing the token, to be reported to user
    :param fq_filename: fully-qualified filename
    :param file_id_ref: unique identifier for current file: [string or None]
    :param unmatched_subtokens: sequence of subtokens, each of which failed
                                spell check
    :param dicts: dictionary set against which to perform matching
    :type  dicts: CorporaFile
    :param ignores: set of tokens to ignore for this session
    :returns: (text, ofs) where ``text`` is the (possibly modified) source
              contents and ``ofs`` is the byte offset within the text where
              searching shall resume.

    """
    token = match_desc.get_token()
    print("%s:%u: Unmatched '%s' --> {%s}" %
          (filename, match_desc.get_line_num(), token,
           ', '.join([st for st in unmatched_subtokens])))
    MATCH_REGEX = re.compile(re.escape(match_desc.get_token()))
    while True:
        print("""\
   (i)gnore, (I)gnore all, (r)eplace, (R)eplace all, (a)dd to dictionary, or
   show (c)ontext? [i]""")
        ch = _portable.getch()
        if ch in (_portable.CTRL_C, _portable.CTRL_D, _portable.CTRL_Z):
            sys.exit(2)
        elif ch in ('i', '\r', '\n'):
            break
        elif ch == 'I':
            ignores.add(token.lower())
            break
        elif ch in ('r', 'R'):
            replacement = raw_input("""\
      Replacement text for '%s': """ % token)
            if replacement == '':
                print("""\
      (Canceled.)\n""")
            else:
                ignores.add(replacement.lower())
                tail = re.sub(
                    MATCH_REGEX, replacement, match_desc.get_remainder(),
                    1 if ch == 'r' else 0)
                print()
                return (match_desc.get_prefix() + tail,
                        match_desc.get_ofs() + len(replacement))
        elif ch == 'a':
            if handle_add(unmatched_subtokens, filename, fq_filename,
                          file_id_ref, dicts):
                break
        elif ch == 'c':
            for ctx in match_desc.get_context():
                print('%4u: %s' % ctx)
            print()
    print()
    # Default: text is unchanged
    return (match_desc.get_string(),
            match_desc.get_ofs() + len(match_desc.get_token()))


def report_failed_check(match_desc, filename, unmatched_subtokens):
    """Handle a token which failed the spell check operation.

    :param match_desc: description of the token matching instance
    :type  match_desc: MatchDescriptor
    :param filename: name of file containing the token
    :param unmatched_subtokens: sequence of subtokens, each of which failed
                                spell check
    :returns: (text, ofs) where ``text`` is the (possibly modified) source
              contents and ``ofs`` is the byte offset within the text where
              searching shall resume.

    """
    token = match_desc.get_token()
    if len(unmatched_subtokens) == 1:
        print(
            "%s:%u: '%s' not found in dictionary (from token '%s')" %
            (filename, match_desc.get_line_num(), unmatched_subtokens[0],
             token),
            file=sys.stderr)
    else:
        unmatched_subtokens = ', '.join(
            "'%s'" %
            t for t in unmatched_subtokens)
        print(
            "%s:%u: %s were not found in the dictionary (from token '%s')" %
            (filename,
             match_desc.get_line_num(),
             unmatched_subtokens,
             token),
            file=sys.stderr)
    # Default: text is unchanged
    return (match_desc.get_string(),
            match_desc.get_ofs() + len(match_desc.get_token()))


def spell_check_token(
        match_desc, filename, fq_filename, file_id_ref,
        dicts, ignores, report_only):
    """Spell check a single token.

    :param match_desc: description of the token matching instance
    :type  match_desc: MatchDescriptor
    :param filename: name of file containing the token, to be reported to user
    :param fq_filename: fully-qualified filename
    :param file_id: unique identifier for this file
    :type  file_id: string or None
    :param dicts: dictionary set against which to perform matching
    :type  dicts: CorporaFile
    :param ignores: set of tokens to ignore for this session
    :returns: ((text, ofs), error_found) where ``text`` is the (possibly
    modified) source contents and ``ofs`` is the byte offset within the text
    where searching shall resume.

    """
    token = match_desc.get_token()
    if (token.lower() not in ignores) and (HEX_REGEX.match(token) is None):
        subtokens = decompose_token(token)
        unmatched_subtokens = [
            st for st in subtokens if len(st) > LEN_THRESHOLD and
            (not dicts.match(st, filename, file_id_ref[0])) and
            (st not in ignores)]
        if unmatched_subtokens:
            unmatched_subtokens = make_unique(unmatched_subtokens)
            if report_only:
                return (report_failed_check(match_desc, filename,
                                            unmatched_subtokens),
                        True)
            else:
                return (
                    handle_failed_check_interactively(
                        match_desc, filename, fq_filename, file_id_ref,
                        unmatched_subtokens, dicts, ignores),
                    True)
    return (
        (match_desc.get_string(), match_desc.get_ofs() + len(token)),
        False)


def spell_check_file(filename, dicts, ignores, report_only, c_escapes):
    """Spell check a single file.

    :param filename: name of the file to check
    :param dicts: dictionary set against which to perform matching
    :type  dicts: CorporaFile
    :param ignores: set of tokens to ignore for this session

    """
    fq_filename = os.path.normcase(os.path.realpath(filename))
    try:
        with _util.open_with_encoding(fq_filename) as source_file:
            source_text = source_file.read()
    except IOError as e:
        print("Error: can't read source file '{}'; "
              'skipping (reason: {})'.format(filename, e),
              file=sys.stderr)
        return False

    # Look for a file ID
    file_id = None
    m_id = FILE_ID_REGEX.search(source_text)
    if m_id is not None:
        file_id = m_id.group(1)
        _util.mutter(
            _util.VERBOSITY_DEBUG,
            '(File contains id "%s".)' %
            file_id)
    else:
        file_id = dicts.file_id_of_file(fq_filename)

    file_id_ref = [file_id]  # allow for spell_check() creating a file_id

    if c_escapes:
        token_regex = C_ESCAPE_TOKEN_REGEX
    else:
        token_regex = TOKEN_REGEX

    # Search for tokens to spell-check
    data = source_text
    pos = 0
    okay = True
    while True:
        m = token_regex.search(data, pos)
        if m is None:
            break
        if (m_id is not None and
                m.start() >= m_id.start() and
                m.start() < m_id.end()):
            # This is matching the file-id.  Skip over it.
            pos = m_id.end()
            continue
        result = spell_check_token(MatchDescriptor(data, m),
                                   filename, fq_filename, file_id_ref,
                                   dicts, ignores, report_only)
        (data, pos) = result[0]
        error_found = result[1]
        if error_found:
            okay = False

    # Write out the source file if it was modified
    if data != source_text:
        with _util.open_with_encoding(fq_filename, mode='w') as source_file:
            try:
                source_file.write(data)
            except IOError as e:
                print(str(e), file=sys.stderr)
                return False

    return okay


def verify_user_data_dir():
    """Verify that the user data directory is present, or create one from
    scratch."""
    if not os.path.exists(USER_DATA_DIR):
        os.makedirs(USER_DATA_DIR)
        shutil.copyfile(SCSPELL_BUILTIN_DICT, DICT_DEFAULT_LOC)


def locate_dictionary():
    """Load the location of the dictionary file.

    This is either the default location, or an override specified in
    'scspell.conf'.

    """
    verify_user_data_dir()
    try:
        f = _util.open_with_encoding(SCSPELL_CONF, encoding='utf-8')
    except IOError:
        return DICT_DEFAULT_LOC

    config = ConfigParser.RawConfigParser()
    try:
        config.readfp(f)
    except ConfigParser.ParsingError as e:
        print(str(e))
        sys.exit(1)
    finally:
        f.close()

    try:
        loc = config.get(CONFIG_SECTION, 'dictionary')
        if os.path.isabs(loc):
            return loc
        else:
            print('Error while parsing "%s": dictionary must be an absolute '
                  'path.' % SCSPELL_CONF)
            sys.exit(1)
    except ConfigParser.Error:
        return DICT_DEFAULT_LOC


def set_dictionary(filename):
    """Set the location of the dictionary to the specified filename.

    :returns: None

    """
    filename = os.path.realpath(
        os.path.expandvars(
            os.path.expanduser(
                filename)))

    verify_user_data_dir()
    config = ConfigParser.RawConfigParser()
    try:
        config.read(SCSPELL_CONF)
    except ConfigParser.ParsingError as e:
        print(str(e))
        sys.exit(1)

    try:
        config.add_section(CONFIG_SECTION)
    except ConfigParser.DuplicateSectionError:
        pass
    config.set(CONFIG_SECTION, 'dictionary', filename)

    with _util.open_with_encoding(SCSPELL_CONF, encoding='utf-8',
                                  mode='w') as f:
        config.write(f)


def export_dictionary(filename, base_dicts):
    """Export the current keyword dictionary to the specified file.

    :returns: None

    """
    if (base_dicts):
        raise SystemExit(
            "--export-dictionary doesn't support " +
            '--base-dict')
        return
    shutil.copyfile(locate_dictionary(), filename)


def find_dict_file(override_dictionary):
    verify_user_data_dir()
    dict_file = locate_dictionary(
    ) if override_dictionary is None else override_dictionary

    return os.path.expandvars(os.path.expanduser(dict_file))


def spell_check(source_filenames, override_dictionary=None,
                base_dicts=[],
                relative_to=None, report_only=False, c_escapes=True,
                test_input=False):
    """Run the interactive spell checker on the set of source_filenames.

    If override_dictionary is provided, it shall be used as a dictionary
    filename for this session only.

    :returns: None

    """
    if test_input:
        _portable.allow_non_terminal_input()

    dict_file = find_dict_file(override_dictionary)

    okay = True
    with CorporaFile(dict_file, base_dicts, relative_to) as dicts:
        ignores = set()
        for f in source_filenames:
            if not spell_check_file(f, dicts, ignores, report_only, c_escapes):
                okay = False
    return okay


def filter_out_base_dicts(override_dictionary=None, base_dicts=[]):
    """Remove from our dictionary the words from the base dicts.

    This can be useful for migrating from a version of scspell that did
    not support the --base-dicts option.

    """
    dict_file = find_dict_file(override_dictionary)
    with CorporaFile(dict_file, base_dicts, None) as dicts:
        dicts.filter_out_base_dicts()


def merge_file_ids(merge_from, merge_to,
                   override_dictionary=None, base_dicts=[], relative_to=None):
    """Merge the file IDs specified by merge_to and merge_from.

    Combine the wordlists in the specified dictionary, and their file ID map
    entries.  They each may be either a file ID, or a filename.  If a filename,
    the file ID corresponding to it is the one merged.

    Use merge_to for the result, discarding merge_from.

    """
    dict_file = find_dict_file(override_dictionary)

    with CorporaFile(dict_file, base_dicts, relative_to) as dicts:
        dicts.merge_file_ids(merge_from, merge_to)


def copy_file(copy_from, copy_to,
              override_dictionary=None, base_dicts=[], relative_to=None):
    """Set up the file copy_to to use the same per-file dictionary as
    copy_from."""
    dict_file = find_dict_file(override_dictionary)

    with CorporaFile(dict_file, base_dicts, relative_to) as dicts:
        dicts.copy_file(copy_from, copy_to)


def rename_file(rename_from, rename_to,
                override_dictionary=None, base_dicts=[], relative_to=None):
    """Rename the file rename_from to rename_to.

    This is with respect to the file ID mappings.

    """
    dict_file = find_dict_file(override_dictionary)

    with CorporaFile(dict_file, base_dicts, relative_to) as dicts:
        dicts.rename_file(rename_from, rename_to)


def add_to_dict(dictionary_type, word, files=[],
                override_dictionary=None, base_dicts=[], relative_to=None):
    """Add word to dictionary_type.

    This is with respect to the filename ID mappings if 'file' type
    dictionary is used."""
    dict_file = find_dict_file(override_dictionary)

    with CorporaFile(dict_file, base_dicts, relative_to) as dicts:
        if dictionary_type[0] == 'n':
            dicts.add_natural(word)

        elif dictionary_type[0] == 'f':
            fq_filename = os.path.normcase(os.path.realpath(files[0]))
            file_id = dicts.file_id_of_file(fq_filename)
            if not file_id:
                file_id = get_new_file_id()
                print('New file ID {0} for {1}'.format(file_id, files[0]),
                      file=sys.stderr)
                dicts.new_file_and_file_id(fq_filename, file_id)
            dicts.add_by_file_id(word, file_id)

        elif dictionary_type[0] == 'p':
            ext = re.sub(r'.*\.', '.', '.{}'.format(files[0].lower()))
            if not dicts.add_by_extension(word, ext):
                print("Dictionary for file extension '{}' not found."
                      .format(ext), file=sys.stderr)

        else:
            print("Dictionary type '{}' not recognized."
                  .format(dictionary_type), file=sys.stderr)


def delete_files(delete_files,
                 override_dictionary=None, base_dicts=[], relative_to=None):
    """Remove all trace of delete_file."""
    dict_file = find_dict_file(override_dictionary)
    with CorporaFile(dict_file, base_dicts, relative_to) as dicts:
        for file in delete_files:
            dicts.delete_file(file)


def main():
    parser = argparse.ArgumentParser(description=__doc__, prog='scspell')

    dict_group = parser.add_argument_group('dictionary file management')
    spell_group = parser.add_argument_group('spell-check control')
    test_group = parser.add_argument_group('testing options')

    spell_group.add_argument(
        '--report-only', dest='report', action='store_true',
        help='non-interactive report of spelling errors')
    spell_group.add_argument(
        '--no-c-escapes', dest='c_escapes',
        action='store_false', default=True,
        help='treat \\label as label, for e.g. LaTeX')

    dict_group.add_argument(
        '--override-dictionary', dest='override_filename',
        help='set location of dictionary to FILE, for current session only',
        metavar='FILE', action='store')
    dict_group.add_argument(
        '--set-dictionary', dest='dictionary',
        help='permanently set location of dictionary to FILE', metavar='FILE',
        action='store')
    dict_group.add_argument(
        '--export-dictionary', dest='export_filename',
        help='export current dictionary to FILE', metavar='FILE',
        action='store')
    dict_group.add_argument(
        '--base-dict', dest='base_dicts', action='append', default=[],
        metavar='BASE_DICT',
        help="Match words from BASE_DICT, but don't modify it.")
    dict_group.add_argument(
        '--use-builtin-base-dict', dest='base_dicts',
        action='append_const', const=SCSPELL_BUILTIN_DICT,
        help="Use scspell's default wordlist as a base dictionary ({0})"
        .format(SCSPELL_BUILTIN_DICT))
    dict_group.add_argument(
        '--filter-out-base-dicts', action='store_true',
        help='Remove from the dictionary file '
             'all the words from the base dicts')
    dict_group.add_argument(
        '--relative-to', dest='relative_to',
        help='use file paths relative to here in file ID map.  '
             'This is required to enable use of the file ID map',
        action='store')
    dict_group.add_argument(
        '-i', '--gen-id', dest='gen_id', action='store_true',
        help='generate a unique file-id string')
    dict_group.add_argument(
        '--merge-file-ids', nargs=2,
        metavar=('FROM_ID', 'TO_ID'),
        help='merge these two file IDs, keeping TO_ID and discarding FROM_ID; '
             'combine their word lists in the dictionary, and the filenames '
             'associated with them in the file ID map; TO_ID and FROM_ID may '
             'be given as file IDs, or as filenames in which case the file '
             'IDs corresponding to those files are operated on; does NOT look '
             'for or consider any file IDs embedded in to-be-spell-checked '
             'files; if your filenames look like file IDs, do it by hand')
    dict_group.add_argument(
        '--copy-file', nargs=2,
        metavar=('FROM_FILE', 'TO_FILE'),
        help='inform scspell that TO_FILE is a copy of FROM_FILE; '
             'effectively, set up TO_FILE to use the same per-file dictionary '
             'as FROM_FILE')
    dict_group.add_argument(
        '--rename-file', nargs=2,
        metavar=('FROM_FILE', 'TO_FILE'),
        help='inform scspell that FROM_FILE has been renamed TO_FILE; '
             'if an entry in the file ID mapping references FROM_FILE, it '
             'will be modified to reference TO_FILE instead')
    dict_group.add_argument(
        '--delete-files', action='store_true', default=False,
        help='inform scspell that the listed files have been deleted; all '
             'file ID mappings for the files will be removed; if all uses of '
             'that file ID have been removed, the corresponding file-private '
             'dictionary will be removed; this will not spell check the '
             'files')
    dict_group.add_argument(
        '--add-to-dict', nargs=2,
        metavar=('DICT_TYPE', 'WORD'),
        help="Add WORD to DICT_TYPE dictionary. If adding to 'file' or "
             "'programming' dictionary then file argument is also required. "
             'Possible DICT_TYPE values are n[atural], p[rogramming], f[ile]')

    #  Testing option to allow scspell to read stdin from a non-tty
    test_group.add_argument(
        '--test-input', action='store_true', default=False,
        help=argparse.SUPPRESS)

    parser.add_argument(
        '-D', '--debug', dest='debug', action='store_true',
        help='print extra debugging information')
    parser.add_argument(
        '--version', action='version',
        version='%(prog)s ' + __version__)
    parser.add_argument(
        'files', nargs='*', help='files to check')

    args = parser.parse_args()

    if args.debug:
        set_verbosity(VERBOSITY_MAX)

    if args.gen_id:
        print('scspell-id: %s' % get_new_file_id())
    elif args.dictionary is not None:
        set_dictionary(args.dictionary)
    elif args.export_filename is not None:
        export_dictionary(args.export_filename, args.base_dicts)
        print("Exported dictionary to '{}'".format(args.export_filename),
              file=sys.stderr)
    elif args.merge_file_ids is not None:
        merge_file_ids(args.merge_file_ids[0], args.merge_file_ids[1],
                       args.override_filename,
                       args.base_dicts, args.relative_to)
    elif args.copy_file is not None:
        copy_file(args.copy_file[0], args.copy_file[1],
                  args.override_filename,
                  args.base_dicts, args.relative_to)
    elif args.rename_file is not None:
        rename_file(args.rename_file[0], args.rename_file[1],
                    args.override_filename,
                    args.base_dicts, args.relative_to)
    elif args.delete_files:
        if len(args.files) < 1:
            parser.error('No files specified for delete')
        delete_files(args.files,
                     args.override_filename,
                     args.base_dicts, args.relative_to)
    elif args.add_to_dict is not None:
        dictionary_type = str(args.add_to_dict[0])
        if dictionary_type in ['p', 'programming'] and len(args.files) < 1:
            parser.error('No file (or extension) specified')
        elif dictionary_type in ['f', 'file'] and len(args.files) < 1:
            parser.error('No file specified')
        elif dictionary_type in ['f', 'file'] and not args.relative_to:
            parser.error("--relative-to is required in order to use '{}' "
                         'dictionary type'.format(dictionary_type))
        elif dictionary_type not in ['p', 'programming',
                                     'f', 'file', 'n', 'natural']:
            parser.error("Dictionary type '{}' not found."
                         .format(dictionary_type))

        add_to_dict(args.add_to_dict[0], args.add_to_dict[1],
                    args.files,
                    args.override_filename,
                    args.base_dicts,
                    args.relative_to)
    elif args.filter_out_base_dicts:
        filter_out_base_dicts(args.override_filename, args.base_dicts)
    elif len(args.files) < 1:
        parser.error('No files specified')
    else:
        okay = spell_check(args.files,
                           args.override_filename,
                           args.base_dicts,
                           args.relative_to,
                           args.report,
                           args.c_escapes,
                           args.test_input)
        return 0 if okay else 1



#
# scspell
# Copyright (C) 2009 Paul Pelzl
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 2, as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

"""Utility functions which may be useful across the source tree."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import io


# Settings for this session
VERBOSITY_NORMAL = 1
VERBOSITY_DEBUG = 2
VERBOSITY_MAX = VERBOSITY_DEBUG
SETTINGS = {'verbosity': VERBOSITY_NORMAL}


def mutter(level, text):
    """Print text to the console, if the level is not higher than the current
    verbosity setting."""
    if level <= SETTINGS['verbosity']:
        print(text)


def set_verbosity(value):
    """Set the verbosity level to a given integral value.

    The constants VERBOSITY_* are good choices.

    """
    SETTINGS['verbosity'] = value


def open_with_encoding(filename, encoding=None, mode='r'):
    """Return opened file with a specific encoding."""
    if not encoding:
        encoding = detect_encoding(filename)

    return io.open(filename, mode=mode, encoding=encoding,
                   newline='')  # Preserve line endings


def detect_encoding(filename):
    """Return file encoding."""

    try:
        input_file = open(filename, 'rb')
    except (IOError, OSError):
        # If the file doesn't exist, return the same thing
        # detect_encoding gives us for an empty file, utf-8.
        return 'utf-8'

    try:
        with input_file:
            from lib2to3.pgen2 import tokenize as lib2to3_tokenize
            encoding = lib2to3_tokenize.detect_encoding(input_file.readline)[0]

            # Check for correctness of encoding.
            with open_with_encoding(filename, encoding) as input_file:
                input_file.read()

        return encoding
    except (SyntaxError, LookupError, UnicodeDecodeError):
        return 'latin-1'



#
# scspell
# Copyright (C) 2009 Paul Pelzl
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 2, as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

"""Contains functions for hiding differences between platforms."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import os
import sys

# Special key codes returned from getch()
CTRL_C = '\x03'
CTRL_D = '\x04'
CTRL_Z = '\x1a'

getch = None
# Cross-platform version of getch()
try:
    import msvcrt

    def msvcrt_getch():
        return msvcrt.getch()
    getch = msvcrt_getch

except ImportError:
    import tty
    import termios

    def termios_getch():
        fd = sys.stdin.fileno()
        old_settings = termios.tcgetattr(fd)
        try:
            tty.setraw(fd)
            ch = sys.stdin.read(1)
        finally:
            termios.tcsetattr(fd, termios.TCSADRAIN, old_settings)
        return ch
    getch = termios_getch


def allow_non_terminal_input():
    def testing_getch():
        s = sys.stdin.read(1)
        if s == '':
            return CTRL_D
        return s
    global getch
    getch = testing_getch


def get_data_dir(prog_name):
    """Retrieves a platform-appropriate data directory for the specified
    program."""
    if sys.platform == 'win32':
        parent_dir = os.getenv('APPDATA')
        prog_dir = prog_name
    else:
        parent_dir = os.path.expanduser('~')
        prog_dir = '.' + prog_name
    return os.path.normpath(os.path.join(parent_dir, prog_dir))



#!/usr/bin/env python

import ast

import setuptools


def get_version():
    """Return version string."""
    with open('scspell/__init__.py') as input_file:
        for line in input_file:
            if line.startswith('__version__'):
                return ast.parse(line).body[0].value.s


with open('README.rst', 'r') as readme_file:
    description = readme_file.read()


setuptools.setup(
    name='scspell3k',
    version=get_version(),
    description='A conservative interactive spell checker for source code.',
    long_description=description,
    url='https://github.com/myint/scspell',
    packages=['scspell'],
    entry_points={
        'console_scripts': ['scspell = scspell:main']},
    package_data={'scspell': ['data/*']},
    license='GPL 2',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License (GPL)',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Topic :: Software Development',
        'Topic :: Text Processing :: Linguistic',
        'Topic :: Utilities'],
    platforms=['any']
)



#
# scspell
# Copyright (C) 2009 Paul Pelzl
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 2, as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

"""Defines methods for storing dictionaries and performing searches against
them."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from collections import OrderedDict
import errno
import io
import json
import os
import re
import sys
from bisect import bisect_left
from . import _util


DICT_TYPE_NATURAL = 'NATURAL'       # Identifies natural language dictionary
DICT_TYPE_FILETYPE = 'FILETYPE'      # Identifies file-type-specific dictionary
DICT_TYPE_FILEID = 'FILEID'        # Identifies file-specific dictionary


# Valid file ID strings take this form
FILE_ID_REGEX = re.compile(r'[a-zA-Z0-9_\-]+')


MATCH_NATURAL = 0x1
MATCH_FILETYPE = 0x2
MATCH_FILEID = 0x4


class ParsingError(Exception):

    """An error occurred when parsing the dictionary file."""


class Corpus(object):

    """Base class for various types of (textual) dictionary-like objects."""

    def __init__(self, dict_type, metadata):
        self._dirty = False
        self._dict_type = dict_type
        self._metadata = metadata

    def _mark_dirty(self):
        self._dirty = True

    def _mark_clean(self):
        self._dirty = False

    def is_dirty(self):
        return self._dirty

    def get_name(self):
        """Get the descriptive name of this dictionary."""
        assert self._dict_type == DICT_TYPE_FILETYPE
        (name, _) = self._metadata
        return name

    def get_extensions(self):
        """Get the list of extensions associated with this dictionary."""
        assert self._dict_type == DICT_TYPE_FILETYPE
        (_, extensions) = self._metadata
        return extensions

    def add_extension(self, extension):
        """Append the extension to the list of extensions associated with this
        dictionary."""
        assert self._dict_type == DICT_TYPE_FILETYPE
        (_, extensions) = self._metadata
        extensions.append(extension)

    def match(self, token):
        """Return True if the token is present in this Corpus.

        The matching method is Corpus-specific.

        """
        raise NotImplementedError

    def add(self, token):
        """Add the specified token to this Corpus."""
        raise NotImplementedError

    def write(self, f):
        """Write the contents of this Corpus to f, a file-like object."""
        raise NotImplementedError

    def _write_header(self, f):
        """Write the corpus header to f, a file-like object."""
        if self._dict_type == DICT_TYPE_NATURAL:
            f.write('%s:\n' % DICT_TYPE_NATURAL)
        elif self._dict_type == DICT_TYPE_FILETYPE:
            (name, extensions) = self._metadata
            f.write(
                '%s: %s; %s\n' % (
                    DICT_TYPE_FILETYPE,
                    name,
                    ', '.join(
                        extensions)))
        elif self._dict_type == DICT_TYPE_FILEID:
            f.write('%s: %s\n' % (DICT_TYPE_FILEID, self._metadata))
        else:
            raise AssertionError('Unknown dict_type "%s".' % self._dict_type)


class ExactMatchCorpus(Corpus):

    """A token matches against an ExactMatchCorpus iff it is present in the
    corpus."""

    def __init__(self, dict_type, metadata, tokens):
        """Construct an instance from a sequence of tokens, giving it the
        specified dictionary type and associated metadata."""
        Corpus.__init__(self, dict_type, metadata)
        self._tokens = set(tokens)

    def match(self, token):
        """Return True if the token is present in this Corpus."""
        return token in self._tokens

    def add(self, token):
        """Add the specified token to this Corpus."""
        if token not in self._tokens:
            self._tokens.add(token)
            self._mark_dirty()

    def write(self, f):
        """Write the contents of this Corpus to f, a file-like object."""
        self._write_header(f)
        for token in sorted(list(self._tokens)):
            f.write(token + '\n')
        f.write('\n')
        self._mark_clean()


class PrefixMatchCorpus(Corpus):

    """A token matches against a PrefixMatchCorpus iff the token is a prefix of
    any item in the corpus."""

    def __init__(self, dict_type, metadata, tokens):
        """Construct an instance from a sequence of tokens, giving it the
        specified dictionary type and associated metadata."""
        Corpus.__init__(self, dict_type, metadata)
        self._tokens = sorted(tokens)

    def match(self, token):
        """Return True if the token is a prefix of an item in this Corpus."""
        insertion_point = bisect_left(self._tokens, token)
        if insertion_point < len(self._tokens):
            return self._tokens[insertion_point].startswith(token)
        else:
            return False

    def add(self, token):
        """Add the specified token to this Corpus."""
        insertion_point = bisect_left(self._tokens, token)
        if (insertion_point >= len(self._tokens) or
                self._tokens[insertion_point] != token):
            self._tokens.insert(insertion_point, token)
            self._mark_dirty()

    def write(self, f):
        """Write the contents of this Corpus to f, a file-like object."""
        self._write_header(f)
        for token in self._tokens:
            f.write(token + '\n')
        f.write('\n')
        self._mark_clean()


class CorporaFile(object):

    """The CorporaFile manages a single file containing multiple corpora.

    May include filename<->file ID mapping file too.

    """

    def __init__(self, filename, base_dicts, relative_to):
        """Construct an instance from the file with the given filename.

        If there are any base_dicts, load them for checking against,
        but don't modify them or write them out.

        relative_to is the directory to consider paths relative to wrt
        the file ID mapping.

        """
        self._base_corpora_files = []
        for fn in base_dicts:
            self._base_corpora_files.append(
                CorporaFile(fn, [], relative_to))

        self._filename = filename

        # Empty defaults
        self._natural_dict = None     # Natural language dictionary
        self._filetype_dicts = []     # File-type-specific dictionaries
        self._file_id_dicts = []      # File-specific dictionaries
        self._extensions = {}
        # Associates each extension with a file-type dictionary
        self._file_ids = {}
        # Associates each file-id with a file-specific dictionary

        self._relative_to = None
        if relative_to is not None:
            self._relative_to = os.path.normcase(os.path.realpath(relative_to))
        self._file_id_mapping = {}
        self._file_id_mapping_is_dirty = False
        # mapping of file ID -> list-of-filenames for file IDs not
        # stored in the source files.
        self._reverse_file_id_mapping = {}
        # Reverse map of the above, individual filename -> file ID

        try:
            with _util.open_with_encoding(filename, mode='r') as f:
                lines = [line.strip(' \r\n') for line in f.readlines()]
            self._parse(lines)
        except IOError as e:
            print(
                'Warning: unable to read dictionary file '
                "'{}' (reason: {})".format(filename, e),
                file=sys.stderr)
        except ParsingError as e:
            raise SystemExit(
                "Error while parsing dictionary file '{}': {}".format(
                    filename, e))

        if self._natural_dict is None:
            print('Continuing with empty natural dictionary\n',
                  file=sys.stderr)
            self._natural_dict = PrefixMatchCorpus(DICT_TYPE_NATURAL, '', [])

        if not self._relative_to:
            return
        mapping_file = self._filename + '.fileids.json'
        try:
            with io.open(mapping_file, mode='r', encoding='utf-8') as mf:
                try:
                    self._file_id_mapping = json.load(mf)
                    _util.mutter(_util.VERBOSITY_DEBUG,
                                 'got file ID mapping:\n{0}'
                                 .format(self._file_id_mapping))
                except ValueError as e:
                    # Error during file creation might leave an empty file
                    # here.  Not necessarily fatal, but report it.
                    _util.mutter(_util.VERBOSITY_NORMAL,
                                 "Couldn't load file ID mapping from {0}: {1}"
                                 .format(mapping_file, e))
        except IOError as e:
            if e.errno == errno.ENOENT:
                _util.mutter(_util.VERBOSITY_DEBUG,
                             'No file ID mappings file {0}'.format(
                                 mapping_file))
            else:
                raise SystemExit(
                    "Can't read file ID mappings file {0}: {1}: {2}".format(
                        mapping_file, e.errno, e.strerror))

        # Build reverse map
        for k, v in self._file_id_mapping.items():
            for f in v:
                self._reverse_file_id_mapping[f] = k

    def match(self, token, filename, file_id,
              match_in=MATCH_NATURAL | MATCH_FILETYPE | MATCH_FILEID):
        """Return True if the token matches any of the applicable corpora.

        :param token: string being matched
        :param filename: name of file containing token
        :param file_id: unique identifier for current file
        :type  file_id: string or None
        :param match_in: Limit the corpora we search
        :returns: True if token matches a dictionary

        """
        for bc in self._base_corpora_files:
            if bc.match(token, filename, file_id, match_in):
                return True

        if match_in & MATCH_NATURAL and self._natural_dict.match(token):
            return True

        if match_in & MATCH_FILETYPE:
            (_, ext) = os.path.splitext(filename.lower())
            try:
                corpus = self._extensions[ext]
                _util.mutter(
                    _util.VERBOSITY_DEBUG,
                    '(Matching against filetype "%s".)' %
                    corpus.get_name())
                if corpus.match(token):
                    return True
            except KeyError:
                _util.mutter(
                    _util.VERBOSITY_DEBUG,
                    '(No filetype match for extension "%s".)' %
                    ext)

        if match_in & MATCH_FILEID and file_id is not None:
            try:
                corpus = self._file_ids[file_id]
                _util.mutter(
                    _util.VERBOSITY_DEBUG,
                    '(Matching against file-id "%s".)' %
                    file_id)
                if corpus.match(token):
                    return True
            except KeyError:
                _util.mutter(
                    _util.VERBOSITY_DEBUG,
                    '(No file-id match for "%s".)' %
                    file_id)

        return False

    def token_is_in_base_dict(self, token, filename, file_id,
                              match_in=MATCH_NATURAL | MATCH_FILETYPE |
                              MATCH_FILEID):
        for bc in self._base_corpora_files:
            if bc.match(token, filename, file_id, match_in):
                return True

    def filter_out_base_dicts(self):
        # For each of our corpora, for each word, if that word is in a
        # base dict, remove it from the corpora.
        #
        # Only remove it when the base dict match was at least as
        # general as the corpora we're processing.  E.g., only remove
        # from our natural_dict when the word was in the natural_dict
        # of some base_dict; not if it was in a filetype or file ID dict.
        # Similarly, only remove from our filetype dict if the word was
        # in a natural_dict or the filetype dict with the same extension.
        new_tokens = []
        for t in self._natural_dict._tokens:
            if self.token_is_in_base_dict(t, None, None, MATCH_NATURAL):
                # Going to change the dict, so mark it dirty
                self._natural_dict._mark_dirty()
            else:
                new_tokens.append(t)
        self._natural_dict._tokens = new_tokens

        for ext in self._extensions:
            # Generate a fake file name to use to query the base dicts.
            # Since we aren't using MATCH_FILEID, the basename won't be
            # used, only the extension.
            fake_filename = 'fake.' + ext
            file_type_corp = self._extensions[ext]
            new_tokens = []
            for t in file_type_corp._tokens:
                if self.token_is_in_base_dict(t, fake_filename, None,
                                              MATCH_NATURAL | MATCH_FILETYPE):
                    file_type_corp._mark_dirty()
                else:
                    new_tokens.append(t)
            file_type_corp._tokens = new_tokens

    def add_natural(self, token):
        """Add the token to the natural language corpus."""
        self._natural_dict.add(token)

    def add_by_extension(self, token, extension):
        """Add the token to a programming language-specific corpus associated
        with the extension.

        Returns True if the add was successful, False if there is no
        corpus with a matching filename extension.

        """
        try:
            corpus = self._extensions[extension]
            _util.mutter(
                _util.VERBOSITY_DEBUG,
                '(Adding to filetype "%s".)' %
                corpus.get_name())
            corpus.add(token)
            return True
        except KeyError:
            _util.mutter(
                _util.VERBOSITY_DEBUG,
                '(No filetype match for extension "%s".)' %
                extension)
            return False

    def add_by_file_id(self, token, file_id):
        """Add the token to a file-specific corpus.

        If there is no corpus for the given file_id, a new one is
        created.

        """
        try:
            corpus = self._file_ids[file_id]
            _util.mutter(
                _util.VERBOSITY_DEBUG,
                '(Adding to file-id "%s".)' %
                file_id)
            corpus.add(token)
        except KeyError:
            _util.mutter(
                _util.VERBOSITY_DEBUG,
                '(No file-id match for "%s"; creating new.)' %
                file_id)
            corpus = ExactMatchCorpus(DICT_TYPE_FILEID, file_id, [])
            self._file_id_dicts.append(corpus)
            self._file_ids[file_id] = corpus
            corpus.add(token)

    def _make_relative_filename(self, fq_filename):
        """return fq_filename relative to self._relative_to."""
        if not fq_filename.startswith(self._relative_to):
            raise SystemExit('File {0} not within --relative-to {1}'.
                             format(fq_filename, self._relative_to))
        rfn = fq_filename[len(self._relative_to):]

        # if relative_to doesn't end in /, we want to make sure we
        # trim the leading / (or /'es) from rfn
        while len(rfn) > 0 and rfn[0] == '/':
            rfn = rfn[1:]
        if len(rfn) == 0:
            raise SystemExit("Making {0} relative to {1}: There's nothing "
                             'left!'.format(fq_filename, self._relative_to))
        return rfn

    def _fn_to_rel(self, filename):
        """Given a filename relative to ".", return the filename relative to
        the --relative-to path."""
        fq_filename = os.path.normcase(os.path.realpath(filename))
        rel_filename = self._make_relative_filename(fq_filename)
        return rel_filename

    def new_file_and_file_id(self, fq_filename, file_id):
        """Add a mapping for this filename and file_id."""

        if self._relative_to is None:
            raise AssertionError('new_file_and_file_id called without '
                                 '--relative-to')
        rel_filename = self._make_relative_filename(fq_filename)
        if rel_filename in self._reverse_file_id_mapping:
            raise AssertionError('{0} already has file_id {1}'.format(
                rel_filename, self._reverse_file_id_mapping[rel_filename]))
        self._reverse_file_id_mapping[rel_filename] = file_id
        if file_id not in self._file_id_mapping:
            self._file_id_mapping[file_id] = []
        if rel_filename not in self._file_id_mapping[file_id]:
            self._file_id_mapping[file_id].append(rel_filename)
            self._file_id_mapping[file_id] = sorted(
                self._file_id_mapping[file_id])
            self._file_id_mapping_is_dirty = True

    def file_id_of_rel_file(self, rel_filename):
        try:
            return self._reverse_file_id_mapping[rel_filename]
        except:
            return None

    def file_id_of_file(self, fq_filename):
        if self._relative_to is None:
            return None
        rel_filename = self._make_relative_filename(fq_filename)
        return self.file_id_of_rel_file(rel_filename)

    def file_id_exists(self, file_id):
        if file_id in self._file_id_mapping:
            return True
        return False

    def merge_file_ids(self, merge_from, merge_to):
        if self.file_id_exists(merge_to):
            id_to = merge_to
        else:
            filename_to = merge_to
            if self._relative_to is not None:
                filename_to = self._fn_to_rel(filename_to)
            id_to = self.file_id_of_rel_file(filename_to)
            if id_to is None:
                raise SystemExit("Can't find merge_to {0} as file ID or file".
                                 format(merge_to))

        if self.file_id_exists(merge_from):
            id_from = merge_from
        else:
            filename_from = merge_from
            if self._relative_to is not None:
                filename_from = self._fn_to_rel(filename_from)
            id_from = self.file_id_of_rel_file(filename_from)
            if id_from is None:
                raise SystemExit("Can't find merge_from {0} as file ID or file"
                                 ''.format(id_from))

        _util.mutter(_util.VERBOSITY_DEBUG,
                     'Going to merge {id_from} into {id_to}'.format(
                         id_from=id_from, id_to=id_to))

        # merge wordlists
        from_corpus = self._file_ids[id_from]
        to_corpus = self._file_ids[id_to]
        for t in from_corpus._tokens:
            to_corpus.add(t)
        del self._file_ids[id_from]
        self._file_id_dicts.remove(from_corpus)

        # Add id_from's files to id_to
        from_files = self._file_id_mapping[id_from]
        to_files = self._file_id_mapping[id_to]
        for f in from_files:
            to_files.append(f)
            self._reverse_file_id_mapping[f] = id_to
        self._file_id_mapping[id_to] = sorted(to_files)
        self._file_id_mapping_is_dirty = True

    def delete_file(self, filename):
        rel_filename = self._fn_to_rel(filename)
        try:
            id = self._reverse_file_id_mapping[rel_filename]
        except:
            if filename == rel_filename:
                report_str = filename
            else:
                report_str = '{0} ({1})'.format(filename, rel_filename)
            _util.mutter(_util.VERBOSITY_NORMAL,
                         'No file ID for {0}'.format(report_str))
            return
        _util.mutter(_util.VERBOSITY_NORMAL,
                     'Removing {0} <-> {1} mappings'.format(
                         filename, id))
        del self._reverse_file_id_mapping[rel_filename]
        fns = self._file_id_mapping[id]
        fns.remove(rel_filename)
        if len(fns) == 0:
            # No remaining files use this file ID.  Remove all trace of it.
            del self._file_id_mapping[id]

            # remove file ID-private dictionary from corpus.
            corpus = self._file_ids[id]
            self._file_id_dicts.remove(corpus)
            del self._file_ids[id]
        self._file_id_mapping_is_dirty = True

    def copy_file(self, copy_from, copy_to):
        from_rel = self._fn_to_rel(copy_from)
        to_rel = self._fn_to_rel(copy_to)
        if from_rel not in self._reverse_file_id_mapping:
            _util.mutter(_util.VERBOSITY_NORMAL,
                         'No file ID for ' + copy_from)
            return

        if to_rel in self._reverse_file_id_mapping:
            self.delete_file(to_rel)

        id_from = self._reverse_file_id_mapping[from_rel]

        _util.mutter(_util.VERBOSITY_NORMAL,
                     'Setting {0} to use {1}\'s file ID {2}.'.format(
                         to_rel, from_rel, id_from))

        fns = self._file_id_mapping[id_from]
        fns.append(to_rel)
        self._file_id_mapping[id_from] = sorted(fns)

        self._reverse_file_id_mapping[to_rel] = id_from
        self._file_id_mapping_is_dirty = True

    def rename_file(self, rename_from, rename_to):
        from_rel = self._fn_to_rel(rename_from)
        to_rel = self._fn_to_rel(rename_to)
        if from_rel not in self._reverse_file_id_mapping:
            _util.mutter(_util.VERBOSITY_NORMAL,
                         'No file ID for ' + rename_from)
            return

        if to_rel in self._reverse_file_id_mapping:
            self.delete_file(to_rel)

        id_from = self._reverse_file_id_mapping[from_rel]

        _util.mutter(_util.VERBOSITY_NORMAL,
                     'Switching file ID {0} from {1} to {2}'.format(
                         id_from, from_rel, to_rel))

        fns = self._file_id_mapping[id_from]
        fns.remove(from_rel)
        fns.append(to_rel)
        self._file_id_mapping[id_from] = sorted(fns)

        self._reverse_file_id_mapping[to_rel] = id_from
        del self._reverse_file_id_mapping[from_rel]
        self._file_id_mapping_is_dirty = True

    def get_filetypes(self):
        """Get a list of file types with type-specific corpora."""
        return [corpus.get_name() for corpus in self._filetype_dicts]

    def new_filetype(self, type_descr, extensions):
        """Add a new file-type corpus with the given description, associated
        with the given set of extensions."""
        assert type_descr not in self.get_filetypes()
        for ext in extensions:
            assert ext not in self._extensions

        corpus = ExactMatchCorpus(
            DICT_TYPE_FILETYPE,
            (type_descr,
             extensions),
            [])
        self._filetype_dicts.append(corpus)
        for ext in extensions:
            self._extensions[ext] = corpus

    def register_extension(self, extension, type_descr):
        """Associate the extension with the file-type that has the given
        description."""
        assert extension not in self._extensions
        for corpus in self._filetype_dicts:
            if corpus.get_name() == type_descr:
                self._extensions[extension] = corpus
                corpus.add_extension(extension)
                return
        raise AssertionError('type_descr "%s" not present.' % type_descr)

    def is_dirty(self):
        dirty = (
            self._natural_dict.is_dirty() if self._natural_dict is not None
            else False
        )
        for corpus in self._filetype_dicts:
            dirty = dirty or corpus.is_dirty()
        for corpus in self._file_id_dicts:
            dirty = dirty or corpus.is_dirty()
        dirty = dirty or self._file_id_mapping_is_dirty
        return dirty

    def close(self):
        """Update the corpus file iff the contents were modified."""
        if self.is_dirty():
            try:
                with _util.open_with_encoding(self._filename, mode='w') as f:
                    for corpus in self._filetype_dicts:
                        corpus.write(f)
                    for corpus in self._file_id_dicts:
                        corpus.write(f)
                    # Natural language dict goes at the end for readability...
                    # it is typically much bigger than the other dictionaries
                    self._natural_dict.write(f)
            except IOError as e:
                print("Warning: unable to write dictionary file '{}' "
                      '(reason: {})'.format(self._filename, e))

        if self._file_id_mapping_is_dirty:
            if self._relative_to is None:
                raise AssertionError('file ID mapping is dirty but ' +
                                     'relative_to is None')

            # Build an OrderedDict sorted by first filename of id, so the
            # mapping file is more reader-friendly.  It will also be
            # more stable, so it will result in less churn if it's checked
            # into git.
            od = OrderedDict()
            copied_ids = set({})
            sorted_filenames = sorted(self._reverse_file_id_mapping)
            for fn in sorted_filenames:
                id = self._reverse_file_id_mapping[fn]
                if id in copied_ids:
                    continue
                copied_ids.add(id)
                od[id] = sorted(self._file_id_mapping[id])

            mapping_file = self._filename + '.fileids.json'
            try:
                with io.open(mapping_file, mode='w', encoding='utf-8') as mf:
                    # http://stackoverflow.com/questions/36003023/json-dump-failing-with-must-be-unicode-not-str-typeerror
                    json_str = json.dumps(od, ensure_ascii=False,
                                          indent=2, separators=(',', ': '))
                    if isinstance(json_str, str):
                        # Apply py2 workaround only on py2
                        if sys.version_info[0] == 2:
                            json_str = json_str.decode('utf-8')
                    mf.write(json_str)
                self._file_id_mapping_is_dirty = False
            except IOError as e:
                print("Warning: unable to write file ID mapping file '{0}' "
                      '(reason: {1})'.format(mapping_file, e))

        # Since we add words only to this, not to any base corpora
        # file, there's nothing to do for the base files now.  But it
        # seems like good form to call close() on them since we've
        # "opened" them.  But be sure we won't actually end up writing
        # any changes out.
        for bc in self._base_corpora_files:
            if bc.is_dirty():
                raise AssertionError('_base_corpora_file is dirty')
            bc.close()

    def _parse(self, lines):
        """Parse the lines into a set of corpora."""
        offset = 0
        while offset < len(lines):
            offset = self._parse_corpus(lines, offset)

    def _parse_corpus(self, lines, offset):
        """Parse a single corpus starting at an offset into lines."""
        (dict_type, metadata) = self._parse_header_line(
            lines[offset], offset + 1)
        (offset, tokens) = _read_corpus_tokens(offset, lines)

        if dict_type == DICT_TYPE_NATURAL:
            self._natural_dict = PrefixMatchCorpus(
                DICT_TYPE_NATURAL, metadata, tokens)
            _util.mutter(
                _util.VERBOSITY_DEBUG,
                '(Loaded natural language dictionary with %u tokens.)' %
                len(tokens))
            return offset

        if dict_type == DICT_TYPE_FILETYPE:
            (type_descr, extensions) = metadata
            corpus = ExactMatchCorpus(DICT_TYPE_FILETYPE, metadata, tokens)
            self._filetype_dicts.append(corpus)
            for ext in extensions:
                self._extensions[ext] = corpus
            _util.mutter(
                _util.VERBOSITY_DEBUG,
                '(Loaded file-type dictionary "%s" with %u tokens.)' %
                (type_descr, len(tokens)))
            return offset

        if dict_type == DICT_TYPE_FILEID:
            corpus = ExactMatchCorpus(DICT_TYPE_FILEID, metadata, tokens)
            self._file_id_dicts.append(corpus)
            self._file_ids[metadata] = corpus
            _util.mutter(
                _util.VERBOSITY_DEBUG,
                '(Loaded file-id dictionary "%s" with %u tokens.)' %
                (metadata, len(tokens)))
            return offset

        raise AssertionError('Unknown dict_type "%s".' % dict_type)

    def _parse_header_line(self, line, line_num):
        """Parse a dictionary header line.

        Headers take the form

            <DICTIONARY TYPE>: <type-specific metadata>

            * If the dictionary type is ``NATURAL`` then the metadata shall be
              empty. The following word list is a natural-language dictionary.

            * If the dictionary type is ``FILETYPE``, then the metadata shall
              take the form "<descriptive name>; <comma-separated extensions
              list>". The name shall be a human-readable description of the
              file type (e.g. the name of the programming language), and the
              extensions list shall be a list of extensions to associate with
              this file type.

            * If the dictionary type is ``FILEID``, then the metadata shall be
              a unique identifier string associated with a particular file

        The return value is the tuple (dictionary type, metadata), where the
        metadata returned is of a type appropriate for the type of dictionary.

        """
        try:
            (raw_dict_type, raw_metadata) = line.split(':')
        except ValueError:
            raise ParsingError('Syntax error in header on line %u.' % line_num)

        dict_type = raw_dict_type.strip()
        metadata = raw_metadata.strip()

        if dict_type == DICT_TYPE_NATURAL:
            if metadata != '':
                raise ParsingError(
                    'Dictionary header "%s" on line %u has nonempty '
                    'metadata.' % (DICT_TYPE_NATURAL, line_num))
            if self._natural_dict is not None:
                raise ParsingError(
                    'Duplicate dictionary type "%s" on line %u.' %
                    (DICT_TYPE_NATURAL, line_num))
            return (dict_type, None)

        if dict_type == DICT_TYPE_FILETYPE:
            try:
                (raw_descr, raw_extensions) = metadata.split(';')
            except ValueError:
                raise ParsingError(
                    'Syntax error in %s dictionary header on line %u.' %
                    (DICT_TYPE_FILETYPE, line_num))

            descr = raw_descr.strip()
            extensions = [
                ext.strip().lower()
                for ext in raw_extensions.split(',')]
            extensions = [ext for ext in extensions if ext != '']

            if descr == '':
                raise ParsingError(
                    'File type-description on line %u is empty.' %
                    line_num)
            for corpus in self._filetype_dicts:
                if corpus.get_name() == descr:
                    raise ParsingError(
                        'Duplicate file-type description "%s" on line %u.' %
                        (descr, line_num))
            if extensions == []:
                raise ParsingError(
                    'Missing extensions list in %s dictionary header on line '
                    '%u.' % (DICT_TYPE_FILETYPE, line_num))
            for ext in extensions:
                if not ext.startswith('.'):
                    raise ParsingError(
                        'Extension "%s" on line %u does not begin with a '
                        'period.' % (ext, line_num))
                if ext in self._extensions:
                    raise ParsingError(
                        'Duplicate extension "%s" on line %u.' %
                        (ext, line_num))
            return (dict_type, (descr, extensions))

        if dict_type == DICT_TYPE_FILEID:
            if FILE_ID_REGEX.match(metadata) is None:
                raise ParsingError(
                    '%s metadata string "%s" on line %u is not a valid file '
                    'ID.' % DICT_TYPE_FILEID, metadata, line_num)
            if metadata in self._file_ids:
                raise ParsingError(
                    'Duplicate file ID string "%s" on line %u.' %
                    (metadata, line_num))
            return (dict_type, metadata)

        raise ParsingError(
            'Unrecognized dictionary type "%s" on line %u.' %
            (dict_type, line_num))

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        self.close()
        return False


def _read_corpus_tokens(offset, lines):
    """Read the set of tokens for the corpus which begins at the given offset.

    Returns the tuple (next offset, tokens).

    """
    tokens = []
    for i, line in enumerate(lines[offset + 1:]):
        if ':' in line:
            return (offset + i + 1, tokens)
        elif line != '':
            tokens.append(line)
    return (len(lines), tokens)
